@@ -329,7 +329,7 @@ _domAssimulate(xmlNodePtr head, xmlNodePtr tail) {
     while ( cur ) {
         /* we must reconcile all nodes in the fragment */
         if (cur->type == XML_DTD_NODE) {
-            if (cur->doc) {
+            if (cur->doc && domGetExternalSubset(cur->doc) != (xmlDtdPtr)cur) {
                 domSetInternalSubset(cur->doc, (xmlDtdPtr) cur);
             }
         }
@@ -355,10 +355,6 @@ _domAssimulate(xmlNodePtr head, xmlNodePtr tail) {
  * the function returns the node inserted. if a fragment was inserted,
  * the first node of the list will returned
  *
- * i ran into a misconception here. there should be a normalization function
- * for the DOM, so sequences of text nodes can get replaced by a single
- * text node. as i see DOM Level 1 does not allow text node sequences, while
- * Level 2 and 3 do.
  **/
 static xmlNodePtr
 _domAddNodeToList(xmlNodePtr cur, xmlNodePtr leader, xmlNodePtr followup, xmlNodePtr *ptail) {
@@ -379,6 +375,10 @@ _domAddNodeToList(xmlNodePtr cur, xmlNodePtr leader, xmlNodePtr followup, xmlNod
             warn("_domAddNodeToList(cur, prev, next, &frag) - 'prev' and 'next' have different parents");
         }
 
+        if (cur->type == XML_DTD_NODE) {
+            xml6_warn("_domAddNodeToList(..) called on a DTD node");
+        }
+
         if ( cur->type == XML_DOCUMENT_FRAG_NODE ) {
             head = _domExtractFrag(cur);
 
@@ -418,6 +418,26 @@ _domAddNodeToList(xmlNodePtr cur, xmlNodePtr leader, xmlNodePtr followup, xmlNod
     return NULL;
 }
 
+static xmlDtdPtr
+_domSetDtd(xmlDocPtr doc, xmlDtdPtr dtd, xmlNodePtr old) {
+    xmlDtdPtr ext_dtd = domGetExternalSubset(doc);
+    int replace_external = (old && old->type == XML_DTD_NODE ? (xmlDtdPtr)old : dtd) == ext_dtd;
+
+    if (doc == NULL) xml6_fail(dtd, "DTD is not associated with a document");
+    if (doc->type != XML_DOCUMENT_NODE && doc->type != XML_HTML_DOCUMENT_NODE && doc->type != XML_DOCB_DOCUMENT_NODE) {
+        xml6_fail((xmlNodePtr)dtd, "appendChild: HIERARCHY_REQUEST_ERR");
+    }
+    if (old) xmlUnlinkNode(old);
+
+    if (replace_external) {
+        domSetExternalSubset(doc, dtd);
+    }
+    else {
+        domSetInternalSubset(doc, dtd);
+    }
+    return dtd;
+}
+
 /**
  * domIsParent tests, if testnode is parent of the reference
  * node. this test is very important to avoid circular constructs in
@@ -771,6 +791,10 @@ domAppendChild( xmlNodePtr self,
         return newChild;
     }
 
+    if ( newChild->type == XML_DTD_NODE ) {
+        _domSetDtd((xmlDocPtr)self, (xmlDtdPtr)newChild, NULL);
+        return newChild;
+    }
     if (self->type == XML_ELEMENT_NODE) {
         if (newChild->type == XML_ATTRIBUTE_NODE) {
             return (xmlNodePtr)domSetAttributeNodeNS(self, (xmlAttrPtr) newChild );
@@ -863,6 +887,10 @@ domReplaceChild( xmlNodePtr self, xmlNodePtr new, xmlNodePtr old ) {
         return domRemoveChild( self, old );
     }
 
+    if (new->type == XML_DTD_NODE) {
+        _domSetDtd((xmlDocPtr)self, (xmlDtdPtr)new, old);
+        return old;
+    }
     if ( old == NULL ) {
         domAppendChild( self, new );
         return old;
@@ -929,6 +957,11 @@ domInsertBefore( xmlNodePtr self,
         return domAppendChild( self, newChild );
     }
 
+    if (newChild->type == XML_DTD_NODE) {
+        _domSetDtd((xmlDocPtr)self, (xmlDtdPtr)newChild, NULL);
+        return newChild;
+    }
+
     if ( !(domTestHierarchy( self, newChild )
            && domTestDocument( self, newChild ))) {
         xml6_fail(self, "insertBefore/insertAfter: HIERARCHY_REQUEST_ERR");
@@ -986,26 +1019,31 @@ domReplaceNode( xmlNodePtr self, xmlNodePtr newNode ) {
         xml6_fail(self, "replaceNode: HIERARCHY_REQUEST_ERR");
     }
 
-    par  = self->parent;
-    prev = self->prev;
-    next = self->next;
-
-    xmlUnlinkNode( self );
-
-    if( prev == NULL && next == NULL ) {
-        /* self was the only child */
-        domAppendChild( par , newNode );
+    if ( newNode->type == XML_DTD_NODE) {
+        _domSetDtd((xmlDocPtr)self->parent, (xmlDtdPtr)newNode, self);
     }
     else {
-        if ( newNode->doc == self->doc ){
-            xmlUnlinkNode( newNode );
+        par  = self->parent;
+        prev = self->prev;
+        next = self->next;
+
+        xmlUnlinkNode( self );
+
+        if (prev == NULL && next == NULL ) {
+            /* self was the only child */
+            domAppendChild( par , newNode );
         }
         else {
-            domImportNode( self->doc, newNode, 1, 0 );
-        }
+            if ( newNode->doc == self->doc ){
+                xmlUnlinkNode( newNode );
+            }
+            else {
+                domImportNode( self->doc, newNode, 1, 0 );
+            }
 
-        head = _domAddNodeToList( newNode, prev,  next, &tail );
-        _domAssimulate(head, tail);
+            head = _domAddNodeToList( newNode, prev,  next, &tail );
+            _domAssimulate(head, tail);
+        }
     }
 
     return self;
@@ -1064,12 +1102,12 @@ domAddSibling( xmlNodePtr self, xmlNodePtr nNode ) {
             xmlFreeNode(copy);
         }
     }
+    else if (nNode->type == XML_DTD_NODE) {
+        _domSetDtd((xmlDocPtr)self->parent, (xmlDtdPtr)nNode, NULL);
+        rv = nNode;
+    }
     else {
         rv = xmlAddSibling( self, nNode );
-
-        if ( rv && nNode->type == XML_DTD_NODE ) {
-            domSetInternalSubset(self->doc, (xmlDtdPtr) nNode);
-        }
     }
     return rv;
 }
